/* pppoatm.c - pppd plugin to implement PPPoATM protocol.
 *
 * Copyright 2000 Mitchell Blank Jr.
 * Based in part on work from Jens Axboe and Paul Mackerras.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */   
#include "pppd.h"
#include "pathnames.h"
#include "fsm.h" /* Needed for lcp.h to include cleanly */
#include "lcp.h"
#include "../../atm2684/atm/lib/atm.h"
#include "../../atm2684/atm/lib/text2atm.h"
#include <linux/atmdev.h>
#include <linux/atmppp.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <net/if.h>
#include <sys/ioctl.h>

static struct sockaddr_atmpvc pvcaddr;
static char *qosstr = NULL;

bool llc_encaps = 0;
bool vc_encaps = 0;
static int device_got_set = 0;
static int pppoatm_max_mtu, pppoatm_max_mru;
extern int     pppd_qos_type;
extern int     pppd_qos_pcr;
extern int     pppd_qos_scr; 
extern int     pppd_qos_mbs;


struct channel pppoatm_channel;
static int fd_pppoa=0;

int setdevname_pppoatm(const char *cp);
static void set_line_discipline_pppoatm(int fd);
static void close_device_pppoatm(void);


/* returns:
 *  -1 if there's a problem with setting the device
 *   0 if we can't parse "cp" as a valid name of a device
 *   1 if "cp" is a reasonable thing to name a device
 * Note that we don't actually open the device at this point
 * We do need to fill in:
 *   devnam: a string representation of the device
 *   devstat: a stat structure of the device.  In this case
 *     we're not opening a device, so we just make sure
 *     to set up S_ISCHR(devstat.st_mode) != 1, so we
 *     don't get confused that we're on stdin.
 */

int setdevname_pppoatm(const char *cp)
{
	struct sockaddr_atmpvc addr;

	if (device_got_set)
		return 0;
	//printf("PPPoATM setdevname_pppoatm\n");
	memset(&addr, 0, sizeof addr);
	if (text2atm(cp, (struct sockaddr *) &addr, sizeof(addr),
	    T2A_PVC | T2A_NAME) < 0)
		return 0;

	memcpy(&pvcaddr, &addr, sizeof pvcaddr);
	strlcpy(devnam, cp, sizeof devnam);

	the_channel = &pppoatm_channel;

	lcp_wantoptions[0].neg_accompression = 0;
	lcp_allowoptions[0].neg_accompression = 0;
	lcp_wantoptions[0].neg_asyncmap = 0;
	lcp_allowoptions[0].neg_asyncmap = 0;
	lcp_wantoptions[0].neg_pcompression = 0;


	//printf("PPPoATM setdevname_pppoatm - SUCCESS\n");
	device_got_set = 1;
	return 1;
}



#define pppoatm_overhead() (llc_encaps ? 6 : 2)

static void no_device_given_pppoatm(void)
{
	fatal("No vpi.vci specified");
}

static int open_device_pppoatm(void)
{
	int fd;
	struct atm_qos qos;

	if (!device_got_set)
		no_device_given_pppoatm();
	fd = socket(AF_ATMPVC, SOCK_DGRAM, 0);
	if (fd < 0)
		printf("failed to create socket: %m");
	
	fd_pppoa = fd;
	memset(&qos, 0, sizeof qos);
	qos.txtp.traffic_class = qos.rxtp.traffic_class = pppd_qos_type;
	/* TODO: support simplified QoS setting */
	if (qosstr != NULL)
		if (text2qos(qosstr, &qos, 0))
			printf("Can't parse QoS: \"%s\"");
	qos.txtp.max_sdu = lcp_allowoptions[0].mru + pppoatm_overhead();
	qos.rxtp.max_sdu = lcp_wantoptions[0].mru + pppoatm_overhead();
	qos.aal = ATM_AAL5;
        qos.txtp.pcr                = pppd_qos_pcr; 
        qos.txtp.scr                = pppd_qos_scr; 
        qos.txtp.mbs                = pppd_qos_mbs; 
	if (setsockopt(fd, SOL_ATM, SO_ATMQOS, &qos, sizeof(qos)) < 0)
		printf("setsockopt(SO_ATMQOS): %m");
	/* TODO: accept on SVCs... */
	if (connect(fd, (struct sockaddr *) &pvcaddr,
	    sizeof(struct sockaddr_atmpvc)))
		printf("connect(%s): %m", devnam);
	pppoatm_max_mtu = lcp_allowoptions[0].mru;
	pppoatm_max_mru = lcp_wantoptions[0].mru;

	set_line_discipline_pppoatm(fd);


	return fd;
}



static void close_device_pppoatm(void)
{

    if (fd_pppoa > 0)
        close(fd_pppoa);
}


static void set_line_discipline_pppoatm(int fd)
{
	struct atm_backend_ppp be;
	be.backend_num = ATM_BACKEND_PPP;
	if (!llc_encaps)
		be.encaps = PPPOATM_ENCAPS_VC;
	else if (!vc_encaps)
		be.encaps = PPPOATM_ENCAPS_LLC;
	else
		be.encaps = PPPOATM_ENCAPS_AUTODETECT;
	if (ioctl(fd, ATM_SETBACKEND, &be) < 0)
		printf("ioctl(ATM_SETBACKEND): %m");
}

static void send_config_pppoatm(int mtu, u_int32_t asyncmap,
	int pcomp, int accomp)
{
	int sock;
	struct ifreq ifr;
	if (mtu > pppoatm_max_mtu)
		printf("Couldn't increase MTU to %d \n", mtu);
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0)
		printf("Couldn't create IP socket: %m");
	strlcpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
	ifr.ifr_mtu = mtu;
	if (ioctl(sock, SIOCSIFMTU, (caddr_t) &ifr) < 0)
		printf("ioctl(SIOCSIFMTU): %m");
	(void) close (sock);
}


static void recv_config_pppoatm(int mru, u_int32_t asyncmap,
	int pcomp, int accomp)
{
	if (mru > pppoatm_max_mru)
		printf("Couldn't increase MRU to %d\n", mru);
}


struct channel pppoatm_channel = {
    options: NULL,
    process_extra_options: NULL,
    check_options: NULL,
    connect: &open_device_pppoatm,
    disconnect: &close_device_pppoatm,
    establish_ppp: &generic_establish_ppp,
    disestablish_ppp: &generic_disestablish_ppp,
    send_config: &send_config_pppoatm,
    recv_config: &recv_config_pppoatm,
    close: NULL,
    cleanup: NULL
};

