/****************************************************************************/

/*
 *      nettel.c -- mappings for NETtel/SecureEdge/SnapGear (x86) boards.
 *
 *      (C) Copyright 2000-2001, Greg Ungerer (gerg@snapgear.com)
 *      (C) Copyright 2001, SnapGear (www.snapgear.com)
 */

/****************************************************************************/

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/cfi.h>
#include <linux/reboot.h>
#include <asm/io.h>

/****************************************************************************/

/*
 *	Define to allow code to support booting directly from any type
 *	of FLASH (INTEL or AMD)...
 */
#if 1
#define	CONFIG_INTEL_BOOT	1
#endif

#define INTEL_BUSWIDTH		1
#define AMD_WINDOW_MAXSIZE	0x00200000
#define AMD_BUSWIDTH	 	1

/*
 *	PAR masks and shifts, assuming 64K pages.
 */
#define SC520_PAR_ADDRESS_MASK	0x00003fff
#define SC520_PAR_ADDRESS_SHIFT	16
#define SC520_PAR_SIZE_MASK	0x01ffc000
#define SC520_PAR_SIZE_SHIFT	2
#define SC520_PAR_ROMCS2	0xca000000	/* Cache disabled, 64K page */
#define SC520_PAR_BOOTCS	0x8a000000

static struct mtd_info *intel_mtd;
static struct mtd_info *amd_mtd;

/****************************************************************************/

static __u8 nettel_read8(struct map_info *map, unsigned long ofs)
{
	return readb(map->map_priv_1 + ofs);
}

static __u16 nettel_read16(struct map_info *map, unsigned long ofs)
{
	return readw(map->map_priv_1 + ofs);
}

static __u32 nettel_read32(struct map_info *map, unsigned long ofs)
{
	return readl(map->map_priv_1 + ofs);
}

static void nettel_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

static void nettel_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	writeb(d, map->map_priv_1 + adr);
}

static void nettel_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	writew(d, map->map_priv_1 + adr);
}

static void nettel_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	writel(d, map->map_priv_1 + adr);
}

static void nettel_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

/****************************************************************************/

static struct map_info nettel_intel_map = {
	name: "NETtel Intel",
	size: 0,
	buswidth: INTEL_BUSWIDTH,
	read8: nettel_read8,
	read16: nettel_read16,
	read32: nettel_read32,
	copy_from: nettel_copy_from,
	write8: nettel_write8,
	write16: nettel_write16,
	write32: nettel_write32,
	copy_to: nettel_copy_to
};

static struct mtd_partition nettel_intel_partitions[] = {
	{
		name: "NETtel kernel",
		offset: 0,
		size: 0x000e0000
	},
	{
		name: "NETtel filesystem",
		offset: 0x00100000,
	},
	{
		name: "NETtel config",
		offset: 0x000e0000,
		size: 0x00020000
	},
	{
		name: "NETtel Intel flash",
		offset: 0
	},
#ifdef CONFIG_INTEL_BOOT
	{
		name: "NETtel first stage boot",
		offset: 0x007e0000,
		size: 0x00020000
	},
#endif
};


static struct map_info nettel_amd_map = {
	name: "NETtel AMD",
	size: AMD_WINDOW_MAXSIZE,
	buswidth: AMD_BUSWIDTH,
	read8: nettel_read8,
	read16: nettel_read16,
	read32: nettel_read32,
	copy_from: nettel_copy_from,
	write8: nettel_write8,
	write16: nettel_write16,
	write32: nettel_write32,
	copy_to: nettel_copy_to
};

static struct mtd_partition nettel_amd_partitions[] = {
	{
		name: "NETtel BIOS config",
		offset: 0x000e0000,
		size: 0x00010000
	},
	{
		name: "NETtel BIOS",
		offset: 0x000f0000,
		size: 0x00010000
	},
	{
		name: "NETtel AMD",
		offset: 0
	},
	{
		name: "NETtel high BIOS",
		offset: 0x001f0000,
		size: 0x00010000
	}
};
#define NUM_AMD_PARTITIONS \
	(sizeof(nettel_amd_partitions)/sizeof(nettel_amd_partitions[0]))

/****************************************************************************/

/*
 *	Set the Intel flash back to read mode since some old boot
 *	loaders don't.
 */
static int nettel_reboot_notifier(
	struct notifier_block *nb,
	unsigned long val,
	void *v)
{
	struct cfi_private *cfi = nettel_intel_map.fldrv_priv;
	
	cfi_send_gen_cmd(0xff, 0x55, 0, &nettel_intel_map, cfi,
		cfi->device_type, NULL);
	return NOTIFY_OK;
}

static struct notifier_block nettel_notifier_block = {nettel_reboot_notifier, NULL, 0 };

/*
 *	Erase the configuration file system.
 *	Used to support the software reset button.
 */
static void nettel_erasecallback(struct erase_info *done)
{
	wait_queue_head_t *wait_q = (wait_queue_head_t *)done->priv;
	wake_up(wait_q);
}

int nettel_eraseconfig(void)
{
	struct mtd_info *mtd;
	struct erase_info erase;
	DECLARE_WAITQUEUE(wait, current);
	wait_queue_head_t wait_q;
	int ret;

	init_waitqueue_head(&wait_q);
	mtd = get_mtd_device(NULL, 2);
	if (mtd) {
		erase.mtd = mtd;
		erase.callback = nettel_erasecallback;
		erase.callback = 0;
		erase.addr = 0;
		erase.len = mtd->size;
		erase.priv = (u_long)&wait_q;
		erase.priv = 0;

		set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&wait_q, &wait);

		ret = MTD_ERASE(mtd, &erase);
		if (ret) {
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&wait_q, &wait);
			put_mtd_device(mtd);
			return ret;
		}

		schedule();  /* Wait for erase to finish. */
		remove_wait_queue(&wait_q, &wait);
		
		put_mtd_device(mtd);
	}

	return 0;
}

/****************************************************************************/

#if LINUX_VERSION_CODE < 0x20212 && defined(MODULE)
#define init_nettel init_module
#define cleanup_nettel cleanup_module
#endif

mod_init_t init_nettel(void)
{
	static void *mmcrp = NULL;
	volatile unsigned long *par13, *par14, *par15;
	unsigned long rom1addr, rom1size;
	unsigned long rom2addr, rom2size;
	unsigned long maxsize, bootaddr;
	int ret, intelboot;
	int num_amd_partitions, num_intel_partitions;

	mmcrp = (void *) ioremap_nocache(0xfffef000, 4096);
	if (mmcrp == NULL) {
		printk("NETtel: failed to disable MMCR cache??\n");
		return -EIO;
	}

	intelboot = 0;
	par14 = (volatile unsigned long *) (mmcrp + 0xc0);
	par13 = (volatile unsigned long *) (mmcrp + 0xbc);

#if CONFIG_INTEL_BOOT
	/*
	 *      This is a quick and dirty hack to support booting from
	 *      Intel FLASH. We should this up a bit, and to allow support
	 *      for 2nd Intel flash for this case too.
	 */
	intelboot = (*par14 == 0);
	if (intelboot)
		par14 = (volatile unsigned long *) (mmcrp + 0xc4);
#endif

	/* Find out where ROMCS1 points to (assuming 64K pages) */
	rom1addr = ((*par14)&SC520_PAR_ADDRESS_MASK)<<SC520_PAR_ADDRESS_SHIFT;

	/* The size register contains the number of pages-1, so add 64k */
	maxsize=0x10000+(((*par14)&SC520_PAR_SIZE_MASK)<<SC520_PAR_SIZE_SHIFT);
	if (maxsize < 0x02000000)
		maxsize = 0x02000000;	/* Detect at least 32MB FLASH chips */

#if 0
	printk(KERN_NOTICE "NETtel: Intel flash device, "
		"address=%lx maxsize=%lx\n", rom1addr, maxsize);
#endif

	/* Set ROMCS1 to the maximum size */
	*par14 = ((*par14) & ~SC520_PAR_SIZE_MASK)
		| (((maxsize-0x10000)>>SC520_PAR_SIZE_SHIFT)&SC520_PAR_SIZE_MASK);

	/* Turn ROMCS2 off so the first probe doesn't find it */
	*par13 = 0;

	/* Probe for the the size of the ROMCS1 flash */
	nettel_intel_map.size = maxsize;
	nettel_intel_map.map_priv_1 = (unsigned long)
		ioremap_nocache(rom1addr, maxsize);
	if (!nettel_intel_map.map_priv_1) {
		printk("NETtel: failed to ioremap() ROMCS1\n");
		return -EIO;
	}
	intel_mtd = do_map_probe("cfi", &nettel_intel_map);
	if (! intel_mtd) {
		iounmap((void *)nettel_intel_map.map_priv_1);
		return -ENXIO;
	}

	rom1size = intel_mtd->size;
#if 0
	printk(KERN_NOTICE "NETtel: Intel flash device, "
		"address=%lx size=%lx\n", rom1addr, rom1size);
#endif

	/* Set ROMCS1 to the detected size */
	*par14 = ((*par14) & ~SC520_PAR_SIZE_MASK)
		| (((rom1size-0x10000)>>SC520_PAR_SIZE_SHIFT)&SC520_PAR_SIZE_MASK);

	/* Map ROMCS2 right after ROMCS1, set its size to the
	 * maximum size that ROMCS1 originally had */
	rom2addr = rom1addr + rom1size;
	*par13 = SC520_PAR_ROMCS2
		| (((maxsize-0x10000)>>SC520_PAR_SIZE_SHIFT) & SC520_PAR_SIZE_MASK)
		| ((rom2addr>>SC520_PAR_ADDRESS_SHIFT) & SC520_PAR_ADDRESS_MASK);
	__asm__ ("wbinvd");

	maxsize += rom1size;

	/* Delete the old map and probe again to do both chips */
	map_destroy(intel_mtd); intel_mtd = NULL;
	iounmap((void *)nettel_intel_map.map_priv_1);

	nettel_intel_map.size = maxsize;
	nettel_intel_map.map_priv_1 = (unsigned long)
		ioremap_nocache(rom1addr, maxsize);
	if (!nettel_intel_map.map_priv_1) {
		printk("NETtel: failed to ioremap() ROMCS1/2\n");
		return -EIO;
	}

	intel_mtd = do_map_probe("cfi", &nettel_intel_map);
	if (! intel_mtd) {
		iounmap((void *)nettel_intel_map.map_priv_1);
		return -ENXIO;
	}

	rom2size = intel_mtd->size - rom1size;
	if (rom2size > 0) {
#if 0
		printk(KERN_NOTICE "NETtel: Intel flash device, "
			"address=%lx size=%lx\n", rom2addr, rom2size);
#endif
		*par13 = SC520_PAR_ROMCS2
			| (((rom2size-0x10000)>>SC520_PAR_SIZE_SHIFT)&SC520_PAR_SIZE_MASK)
			| ((rom2addr>>SC520_PAR_ADDRESS_SHIFT)&SC520_PAR_ADDRESS_MASK);
		__asm__ ("wbinvd");
	} else {
		*par13 = 0;
	}

	printk(KERN_NOTICE "NETtel: Intel flash device size = %dK\n",
		intel_mtd->size>>10);

	intel_mtd->module = THIS_MODULE;
#ifndef CONFIG_BLK_DEV_INITRD
	ROOT_DEV = MKDEV(MTD_BLOCK_MAJOR, 1);
#endif

	num_intel_partitions = sizeof(nettel_intel_partitions) /
		sizeof(nettel_intel_partitions[0]);
#if CONFIG_INTEL_BOOT
	if (intelboot) {
		/* Adjust offset and size of last boot partition */
		nettel_intel_partitions[4].offset = intel_mtd->size -
			intel_mtd->erasesize;
		nettel_intel_partitions[1].size = intel_mtd->size -
			(1024*1024 + intel_mtd->erasesize);
	} else {
		/* No Intel first stage boot when booting from AMD FLASH */
		num_intel_partitions--;
	}
#endif
	ret = add_mtd_partitions(intel_mtd, nettel_intel_partitions,
		num_intel_partitions);
	if (ret != 0)
		return ret;

#if CONFIG_INTEL_BOOT
        if (intelboot)
                return(0);
#endif

	/*
	 *	Disable caching in AMD flash and remap its base address
	 *	to after the Intel flash.
	 */
	bootaddr = rom2addr + rom2size;
	maxsize = AMD_WINDOW_MAXSIZE;
#if 0
	printk(KERN_NOTICE" NETtel: AMD flash device, address=%lx "
		"maxsize=%lx\n", bootaddr, maxsize);
#endif

	par15 = (volatile unsigned long *) (mmcrp + 0xc4);
	*par15 = SC520_PAR_BOOTCS
		| (((maxsize-0x10000)>>SC520_PAR_SIZE_SHIFT)&SC520_PAR_SIZE_MASK)
		| ((bootaddr>>SC520_PAR_ADDRESS_SHIFT)&SC520_PAR_ADDRESS_MASK);
	__asm__ ("wbinvd");

	nettel_amd_map.map_priv_1 = (unsigned long)
		ioremap_nocache(bootaddr, maxsize);
	if (!nettel_amd_map.map_priv_1) {
		printk("NETtel: failed to ioremap() BOOTCS\n");
		return -EIO;
	}
#if 0
	printk(KERN_NOTICE "NETtel: AMD flash device, virtual address=%lx\n",
		nettel_amd_map.map_priv_1);
#endif

	amd_mtd = do_map_probe("cfi", &nettel_amd_map);
	if (! amd_mtd) {
		iounmap((void *)nettel_amd_map.map_priv_1);
		return -ENXIO;
	}

	printk(KERN_NOTICE "NETtel: AMD flash device size = %dK\n",
			amd_mtd->size>>10);

	amd_mtd->module = THIS_MODULE;

	/* The high BIOS partition is only present for 2MB units */
	num_amd_partitions = NUM_AMD_PARTITIONS;
	if (amd_mtd->size < AMD_WINDOW_MAXSIZE) {
		num_amd_partitions--;
	}
	ret = add_mtd_partitions(amd_mtd, nettel_amd_partitions, num_amd_partitions);

	register_reboot_notifier(&nettel_notifier_block);

	return ret;
}

/****************************************************************************/

mod_exit_t cleanup_nettel(void)
{
	unregister_reboot_notifier(&nettel_notifier_block);

	if (amd_mtd) {
		del_mtd_partitions(amd_mtd);
		map_destroy(amd_mtd);
	}
	if (nettel_amd_map.map_priv_1) {
		iounmap((void *)nettel_amd_map.map_priv_1);
		nettel_amd_map.map_priv_1 = 0;
	}
	if (intel_mtd) {
		del_mtd_partitions(intel_mtd);
		map_destroy(intel_mtd);
	}
	if (nettel_intel_map.map_priv_1) {
		iounmap((void *)nettel_intel_map.map_priv_1);
		nettel_intel_map.map_priv_1 = 0;
	}
}

module_init(init_nettel);
module_exit(cleanup_nettel);

/****************************************************************************/
